"""
Store the rotated mass-weighted forces and inertia-weighted torques in this class.
"""

import numpy as np

from waterEntropy.utils.helpers import nested_dict


class CovarianceCollection:
    """
    Class for covariance matrices molecules, grouped by what their closest
    non-like molecule is.
    """

    def __init__(self):
        self.forces = nested_dict()
        self.torques = nested_dict()
        self.counts = nested_dict()

    def add_data(
        self, nearest, molecule_name: str, force: np.ndarray, torque: np.ndarray
    ):
        """
        Add force, torque covariances to the class dictionaries

        :param nearest: the value for grouping molecules into a class. e.g. the
            name of the nearest molecule
        :param molecule_name: the name of the molecule being populated into the
            dictionaries
        :param force: array of the mass-weighted forces of the molecule
        :param torque: array of the inertia-weighted torques of the molecule
        """
        # order is important, update count last
        CovarianceCollection.populate_dicts(
            self, nearest, molecule_name, force, self.forces, self.counts
        )
        CovarianceCollection.populate_dicts(
            self, nearest, molecule_name, torque, self.torques, self.counts
        )
        CovarianceCollection.update_counts(self, nearest, molecule_name, self.counts)

    def populate_dicts(
        self, nearest, molecule_name, variable, variable_dict, count_dict
    ):
        # pylint: disable=too-many-arguments
        """
        For a given molecule, append the summed, weighted and rotated forces,
        and the torques for the whole molecule. Add as a running average by
        keeping count of the number of molecules added.

        :param self: class instance
        :param nearest: the value for grouping molecules into a class. e.g. the
            name of the nearest molecule
        :param molecule_name: name of molecule
        :param variable: variable to update a dict
        :param variable_dict: the dictionary where updated variables are added
        :param count_dict: the dictionary where counts are updated
        """
        # add molecule name to dicts if it doesn't exist
        if (nearest, molecule_name) not in variable_dict:
            variable_dict[(nearest, molecule_name)] = variable
        else:
            # get running average of forces and torques
            stored_variable = variable_dict[(nearest, molecule_name)]
            stored_count = count_dict[(nearest, molecule_name)]
            updated_variable = (stored_variable * stored_count + variable) / (
                stored_count + 1
            )
            # update dictionaries with running averages
            variable_dict[(nearest, molecule_name)] = updated_variable

    def update_counts(self, nearest, molecule_name, count_dict):
        """Update the counts in dictionary

        :param nearest: the value for grouping molecules into a class. e.g. the
            name of the nearest molecule
        :param molecule_name: name of molecule
        :param count_dict: the dictionary where counts are updated
        """
        if (nearest, molecule_name) not in count_dict:
            count_dict[(nearest, molecule_name)] = 1
        else:
            count_dict[(nearest, molecule_name)] += 1



"""
These functions calculate hydrogen bonding within a shell.
"""

import numpy as np

import waterEntropy.analysis.RAD as RADShell
import waterEntropy.maths.trig as Trig
from waterEntropy.utils.helpers import nested_dict


class HBCollection:
    """
    Class for hydrogen bond donors and their acceptors
    """

    def __init__(self):
        self.donating_to = (
            nested_dict()
        )  # structure: UA_idx[donating_idx] = accepting_idx
        self.accepting_from = (
            nested_dict()
        )  # structure: UA_idx[accepting_idx] = [donating_idx1,]

    def add_data(self, UA_idx: int, donator_idx: int, acceptor_idx: int):
        """
        For a given donator index, save its acceptor index in the donating_to
        dictionary. And for that acceptor, save the donator index as a list in
        the accepting from dictionary.

        :param self: class instance
        :param UA_idx: atom index of the UA of the bonded donating hydrogen
        :param donator_idx: atom index of the donator bonded to the UA
        :param acceptor_idx: atom index of the UA accepting the
            hydrogen bond in the coordination shell
        """
        if donator_idx not in self.donating_to[UA_idx]:
            self.donating_to[UA_idx][donator_idx] = acceptor_idx
        if acceptor_idx not in self.accepting_from:
            self.accepting_from[acceptor_idx] = []
        if donator_idx not in self.accepting_from[acceptor_idx]:
            self.accepting_from[acceptor_idx].append(donator_idx)

    def find_donators(self, UA_idx: int):
        """
        Find the donators for a given residue, this returns a dict where the
        key is the accepting atom index and the value is a list of the atom
        indices that donate to it.

        :param cls: class instance
        :param UA_idx: atom index of UA being donated to
        """
        return self.accepting_from.get(UA_idx, None)

    def find_acceptor(self, UA_idx: int):
        """
        Find the acceptors for a given resid, this returns a dictionary where
        the key is the donator index and the value is the acceptor index

        :param cls: class instance
        :param UA_idx: atom index of UA accepting HBs
        """
        return self.donating_to.get(UA_idx, None)


def get_HB_terms(heavy_atom, donator, acceptor, DA_distance, dimensions):
    """
    Get two terms for calculating the hydrogen bond between a donator and
    potential acceptor in a coordination shell.
    For a hydrogen bond to form, the following criteria need to be met:

    1. Angle between heavy atom bonded to donator (X), the donator (D) and
    the acceptor (A) is over 90 degrees

    2. The relative charge is most negative over all other neighbours in a
    coordination shell

    :param heavy_atom: the mdanalysis instance for the heavy atom bonded to
        the donator
    :param donator: the mdanalysis instance for the donator
    :param acceptor: the mdanalysis instance for the possible acceptor
    :param DA_distance: the distance between donor and acceptor
    :param dimensions: the dimensions of the simulation box
    """
    relative_charge = (float(donator.charge) * float(acceptor.charge)) / float(
        DA_distance**2
    )
    cosine_angle = Trig.get_angle(
        heavy_atom.position, donator.position, acceptor.position, dimensions
    )
    XDA_angle = np.degrees(np.arccos(cosine_angle))

    return relative_charge, XDA_angle


def get_shell_HB_acceptors(shell, system, HBs: HBCollection):
    # pylint: disable=too-many-locals
    """
    Find the hydrogen bond acceptors for the central UA hydrogens that are
    electropositive. The assumption is made that hydrogen bonding can only
    occur inside the coordination shell, this is needed for shell neighbour
    labelling used in orientational entropy calculations

    :param shell: the instance for containing coordination shell neighbours
    :param system: mdanalysis instance of all atoms in current frame
    :param HBs: instance of HBCollection class
    """
    X_idx = shell.atom_idx  # shell central heavy atom
    heavy_atom = system.atoms[X_idx]
    # 1. iterate over atoms bonded to atom X
    for i in heavy_atom.bonds:
        D_idx = i.indices[1]
        bonded = system.atoms[D_idx]
        # 2. find hydrogen atoms bonded to heavy atom with positive charge
        if bonded.mass < 1.1 and bonded.mass > 1.0 and bonded.charge > 0:
            donator = system.atoms[D_idx]
            # 3. set starting hydrogen bond term and possible acceptor to a large number
            # the acceptor that gives the lowest relative charge is the acceptor
            current_relative_charge = 99
            current_acceptor = False
            # 4. get shell neighbours ordered by ascending distance
            sorted_indices, sorted_distances = Trig.get_shell_neighbour_selection(
                shell, donator, system
            )
            for A_idx, DA_distance in zip(sorted_indices, sorted_distances):
                # 5. check if neighbouring atom in shell is an acceptor,
                # if so override current possible acceptor
                acceptor = system.atoms[A_idx]
                if not current_acceptor:
                    current_acceptor = acceptor
                relative_charge, XDA_angle = get_HB_terms(
                    heavy_atom, donator, acceptor, DA_distance, system.dimensions[:3]
                )
                # 6. Check if an atom is a possible hydrogen bond acceptor
                if relative_charge < current_relative_charge and float(XDA_angle) > 90:
                    current_relative_charge = relative_charge
                    current_acceptor = acceptor
            # 7. create a new object for hydrogen bonding in a RAD shell
            HBs.add_data(X_idx, D_idx, current_acceptor.index)


def get_shell_HBs(shell, system, HBs: HBCollection, shells: RADShell.ShellCollection):
    """
    For a given UA and its coordination shell neighbours, find what the central
    UA donates to and accepts from in its shell.

    :param shell: the instance for class waterEntropy.neighbours.RAD.RAD
        containing coordination shell neighbours
    :param system: mdanalysis instance of all atoms in current frame
    :param HBs: HBCollection instance
    :param shells: ShellCollection instance
    """
    # 1. first check that HB donations haven't already been found
    donates_to = HBs.find_acceptor(shell.atom_idx)
    if not donates_to:
        get_shell_HB_acceptors(shell, system, HBs)
        donates_to = HBs.find_acceptor(shell.atom_idx)
    # 2. now iterate through shell and find shells of shell neighbours
    for n_idx in shell.UA_shell:
        neighbour_shell = shells.find_shell(n_idx)
        if not neighbour_shell:
            neighbour_shell = RADShell.get_RAD_shell(
                system.atoms[n_idx], system, shells
            )
        # 3. find what each shell neighbour donates to in their shell
        neighbour_donates_to = HBs.find_acceptor(n_idx)
        if not neighbour_donates_to:
            get_shell_HB_acceptors(neighbour_shell, system, HBs)
            neighbour_donates_to = HBs.find_acceptor(n_idx)



"""Store shell information here"""


class Shell:
    """Represent a single atom shell with dynamic properties, which are set in
    the ShellCollection class."""

    def __init__(self, atom_idx):
        object.__setattr__(self, "atom_idx", atom_idx)  # Store atom index directly
        object.__setattr__(self, "properties", {})  # Internal dictionary

    def __getattr__(self, key: str):
        """Allow dot notation access to stored properties.

        :param key: name of the property in the properties dictionary
        """
        if key in self.properties:
            return self.properties[key]
        raise AttributeError(f"Property '{key}' not found for atom {self.atom_idx}")

    def __setattr__(self, key: str, value):
        """Allow adding properties dynamically, except for 'atom_idx'.

        :param key: name of the property being added to the properties dictionary
        :param value: value for the key being added to the properties dictionary
        """
        if key == "atom_idx":  # Ensure atom index remains immutable
            object.__setattr__(self, key, value)
        else:
            self.properties[key] = value  # Store additional properties

    def __repr__(self):
        """Print the atom and its properties."""
        return f"Shell(atom_idx={self.atom_idx}, properties={self.properties})"


class ShellCollection:
    """Manage multiple atom shells and allow adding/updating properties."""

    def __init__(self):
        self.shells = {}  # Dictionary storing Shell objects indexed by atom index

    def add_data(self, atom_idx: int, UA_shell: list[int]):
        """Add a new atom shell to the collection if it doesn't exist and set
        various properties used to describe the class.

        :param atom_idx: the heavy atom index of the central atom in a shell
        :param UA_shell: the list of heavy atom indices in the shell of atom_idx
        """
        if atom_idx not in self.shells:
            self.shells[atom_idx] = Shell(atom_idx)
            self.set_property(atom_idx, "atom_idx", atom_idx)
            self.set_property(atom_idx, "UA_shell", UA_shell)
            self.set_property(atom_idx, "nearest_nonlike_idx", None)
            self.set_property(atom_idx, "labels", None)
            self.set_property(atom_idx, "donates_to_labels", None)
            self.set_property(atom_idx, "accepts_from_labels", None)

    def set_property(self, atom_idx, key: str, value):
        """Set a property for a specific atom shell, ensuring it exists first.

        :param atom_idx: the heavy atom index of the central atom in a shell
        :param key: the name of the key being added to the shells dictionary
        :param value: the value of the key being added to the shells dictionary
        """
        if atom_idx not in self.shells:
            self.shells[atom_idx] = Shell(atom_idx)  # Auto-create atom if missing
        setattr(self.shells[atom_idx], key, value)

    def find_shell(self, atom_idx: int):
        """
        Get the shell instance if atom_idx is a key in shells dictionary

        :param atom_idx: index of central atom in shell
        """
        return self.shells.get(atom_idx, None)

    def __repr__(self):
        """Return a dictionary-like representation of stored atoms."""
        return repr(self.shells)



"""
These functions calculate coordination shells using RAD the relative
angular distance.
"""

import numpy as np

from waterEntropy.analysis.shells import ShellCollection
import waterEntropy.maths.trig as Trig


def get_RAD_neighbours(i_coords, sorted_indices, sorted_distances, system):
    # pylint: disable=too-many-locals
    r"""
    For a given set of atom coordinates, find its RAD shell from the distance
    sorted atom list, truncated to the closests 30 atoms.

    This function calculates coordination shells using RAD the relative
    angular distance, as defined first in DOI:10.1063/1.4961439
    where united atoms (heavy atom + bonded Hydrogens) are defined as neighbours if
    they fulfil the following condition:

    .. math::
        \Bigg(\frac{1}{r_{ij}}\Bigg)^2>\Bigg(\frac{1}{r_{ik}}\Bigg)^2 \cos \theta_{jik}

    For a given particle :math:`i`, neighbour :math:`j` is in its coordination
    shell if :math:`k` is not blocking particle :math:`j`. In this implementation
    of RAD, we enforce symmetry, whereby neighbouring particles must be in each
    others coordination shells.

    :param i_coords: xyz coordinates of atom :math:`i`
    :param sorted_indices: list of atom indices sorted from closest to
        furthest from atom :math:`i`
    :param sorted_distances: list of atom distances sorted from closest to
        furthest from atom :math:`i`
    :param system: mdanalysis instance of atoms in a frame
    """
    # 1. truncate neighbour list to closest 30 united atoms
    range_limit = min(len(sorted_distances), 30)
    shell = []
    count = -1
    # 2. iterate through neighbours from closest to furthest
    for y in sorted_indices[:range_limit]:
        count += 1
        y_idx = np.where(sorted_indices == y)[0][0]
        j = system.atoms.indices[y]
        j_coords = system.atoms.positions[y]
        rij = sorted_distances[y_idx]
        blocked = False
        # 3. iterate through neighbours than atom j and check if they block
        # it from atom i
        for z in sorted_indices[:count]:  # only closer UAs can block
            z_idx = np.where(sorted_indices == z)[0][0]
            k_coords = system.atoms.positions[z]
            rik = sorted_distances[z_idx]
            # 4. find the angle jik
            costheta_jik = Trig.get_angle(
                j_coords, i_coords, k_coords, system.dimensions[:3]
            )
            if np.isnan(costheta_jik):
                break
            # 5. check if k blocks j from i
            LHS = (1 / rij) ** 2
            RHS = ((1 / rik) ** 2) * costheta_jik
            if LHS < RHS:
                blocked = True
                break
        # 6. if j is not blocked from i by k, then its in i's shell
        if blocked is False:
            shell.append(j)
    return shell


def get_RAD_shell(UA, system, shells: ShellCollection):
    """
    For a given united atom, find its RAD shell, returning the atom indices
    for the heavy atoms that are in its shell.

    :param UA: mdanalysis instance of a united atom in a frame
    :param system: mdanalysis instance of atoms in a frame
    :param shells: ShellCollection instance
    """
    # 1. first check if a shell has already been found for this UA
    shell = shells.find_shell(UA.index)
    if not shell:
        # 2. get the nearest neighbours for the UA, sorted from closest to
        # furthest
        sorted_indices, sorted_distances = Trig.get_sorted_neighbours(UA.index, system)
        # 3. now find the RAD shell of the UA
        shell_indices = get_RAD_neighbours(
            UA.position, sorted_indices, sorted_distances, system
        )
        # 4. populate the class instance for RAD shells
        shells.add_data(UA.index, shell_indices)
        shell = shells.find_shell(UA.index)
    return shell



"""
Label neighbours in a coordination shell based on what they are and what their
neighbours are.
"""

import waterEntropy.analysis.RAD as RADShell
from waterEntropy.analysis.shells import ShellCollection


def get_shell_labels(atom_idx: int, system, shell, shells: ShellCollection):
    """
    Get the shell labels of an atoms shell based on the following:
    For a central UA, rank its coordination shell by proximity to that
    central UA's nearest non-like molecule UA.

    * '#_RESNAME' = RAD shell from same molecule type, when nearest nonlike resid is the same as the reference.

    * 'X_RESNAME' = when same molecule type has different nearest nonlike resid.

    * 'RESNAME' = when molecule of different type is in RAD shell.

    * '0_RESNAME' = closest different type molecule in RAD shell. (the one its assigned to, its nearest non-like!)

    :param atom_idx: atom index of central atom in coordination shell
    :param system: mdanalysis instance of atoms in a frame
    :param shell: shell instance of atom_idx
    :param shells: ShellCollection instance
    """
    center = system.atoms[atom_idx]
    # 1. find the closest different UA in a shell
    #   different = not the same resname
    nearest_nonlike_idx = get_nearest_nonlike(shell, system)
    # 2. only find labels if a solute is in the shell
    if nearest_nonlike_idx is not None:
        nearest_nonlike = system.atoms[nearest_nonlike_idx]
        shell_labels = []
        for n in shell.UA_shell:
            neighbour = system.atoms[n]
            # 3a. label nearest nonlike atom as "0_RESNAME"
            if neighbour.index == nearest_nonlike.index:
                shell_labels.append(f"0_{neighbour.resname}")
            # 3b. label other nonlike atoms as "RESNAME"
            if (
                neighbour.index != nearest_nonlike.index
                and neighbour.resname != center.resname
            ):
                shell_labels.append(neighbour.resname)
            # 3c. find RAD shells for shell constituents with same resname
            # as central atom
            if (
                neighbour.index != nearest_nonlike.index
                and neighbour.resname == center.resname
            ):
                neighbour_shell = shells.find_shell(neighbour.index)
                if not neighbour_shell:
                    neighbour_shell = RADShell.get_RAD_shell(neighbour, system, shells)
                # 3d. find nearest nonlike of neighbours with same resname
                # as central atom
                neighbour_nearest_nonlike_idx = get_nearest_nonlike(
                    neighbour_shell, system
                )
                # 3e. if neighbour has a pure shell, then it is in the second
                # shell of the nearest nonlike
                if neighbour_nearest_nonlike_idx is None:
                    shell_labels.append(f"2_{neighbour.resname}")
                else:
                    # 3f. if neighbours nearest nonlike is the same atom as
                    # central atom, assume it is in the first shell
                    # if neighbour_nearest_nonlike_idx == nearest_nonlike_idx:
                    neighbour_nearest_nonlike = system.atoms[
                        neighbour_nearest_nonlike_idx
                    ]
                    # 3g. if neighbours nearest nonlike is in the same resid as
                    # central atom, assume it is in the first shell
                    if neighbour_nearest_nonlike.resid == nearest_nonlike.resid:
                        shell_labels.append(f"1_{neighbour.resname}")
                    else:
                        # 3h. if neighbours nearest nonlike is not the same resid
                        # as central nearest resid,  it is in the first shell
                        # of a different resid and labelled as "X_RESNAME"
                        shell_labels.append(f"X_{neighbour.resname}")
        shell.labels = shell_labels  # sorted(shell_labels) #don't sort yet
        shell.nearest_nonlike_idx = nearest_nonlike.index
    return shell


def get_nearest_nonlike(shell, system):
    """
    For a given shell, find the closest neighbour that is not the same
    atom/molecule type as the central united atom.

    :param shell: shell instance of an atom
    :param system: mdanalysis instance of atoms in a frame
    """
    nearest_nonlike_idx = None
    center = system.atoms[shell.atom_idx]
    for n in shell.UA_shell:
        neighbour = system.atoms[n]
        if neighbour.resname != center.resname and neighbour.type != center.type:
            nearest_nonlike_idx = n
            break
    return nearest_nonlike_idx



#!/usr/bin/env python

"""
"""

import argparse
from datetime import datetime
import logging
import sys

from MDAnalysis import Universe

import waterEntropy.recipes.interfacial_solvent as GetSolvent
import waterEntropy.entropy.vibrations as VIB
import waterEntropy.entropy.orientations as OR

def run_waterEntropy(
    file_topology="file_topology",
    file_coords="file_coords",
    file_forces="file_forces",
    file_energies="file_energies",
    list_files="list_files",
    start="start",
    end="end",
    step="step",
):
    # pylint: disable=all
    """
    Functions required:
    [X] Force torque for whole molecule principal axes
    [X] Force torque for custom axes
    [X] Covariance matrices
    [X] Nearest non-like
    [X] RAD shell
    [X] HB in shell
    [X] neighbour list
    [X] label RAD shell
    [] molecule level RAD shell
    [] save stats into multiple dicts
    [] running averages
    [] Dict for
    """

    startTime = datetime.now()
    print(startTime)

    # load topology and coordinates
    u = Universe(file_topology, file_coords, format="MDCRD")
    # seperate universe where forces are the loaded trajectory
    uf = Universe(file_topology, file_forces, format="MDCRD")
    # set the has_forces flag on the Timestep first
    u.trajectory.ts.has_forces = True
    # add the forces (which are saved as positions be default) from uf to u
    u.atoms.forces = uf.atoms.positions
    # set the frames to be analysed
    #start, end, step = 0, 4, 2
    print(u.trajectory)
    # u.trajectory[frame] # move to a particular frame using this

    Sorient_dict, covariances, vibrations, frame_solvent_indices = GetSolvent.get_interfacial_water_orient_entropy(u, start, end, step)
    OR.print_Sorient_dicts(Sorient_dict)
    # GetSolvent.print_frame_solvent_dicts(frame_solvent_indices)
    VIB.print_Svib_data(vibrations, covariances)


    sys.stdout.flush()
    print(datetime.now() - startTime)


def main():
    """ """
    try:
        usage = "runWaterEntropy.py [-h]"
        parser = argparse.ArgumentParser(
            description="Program for reading "
            "in molecule forces, coordinates and energies for "
            "entropy calculations.",
            usage=usage,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        parser.add_argument_group("Options")
        parser.add_argument(
            "-top",
            "--file_topology",
            metavar="file",
            default=None,
            help="name of file containing system topology.",
        )
        parser.add_argument(
            "-crd",
            "--file_coords",
            metavar="file",
            default=None,
            help="name of file containing coordinates.",
        )
        parser.add_argument(
            "-frc",
            "--file_forces",
            metavar="file",
            default=None,
            help="name of file containing forces.",
        )
        parser.add_argument(
            "-ener",
            "--file_energies",
            metavar="file",
            default=None,
            help="name of file containing energies.",
        )
        parser.add_argument(
            "-l",
            "--list_files",
            action="store",
            metavar="file",
            default=False,
            help="file containing list of file paths.",
        )
        parser.add_argument(
            "-s",
            "--start",
            action="store",
            metavar="int",
            default=0,
            help="frame number to start analysis from.",
        )
        parser.add_argument(
            "-e",
            "--end",
            action="store",
            metavar="int",
            default=1,
            help="frame number to end analysis at.",
        )
        parser.add_argument(
            "-dt",
            "--step",
            action="store",
            metavar="int",
            default=1,
            help="steps to take between start and end frame selections.",
        )
        op = parser.parse_args()
    except argparse.ArgumentError:
        logging.error(
            "Command line arguments are ill-defined, please check the arguments."
        )
        raise
        sys.exit(1)

    run_waterEntropy(
        file_topology=op.file_topology,
        file_coords=op.file_coords,
        file_forces=op.file_forces,
        file_energies=op.file_energies,
        list_files=op.list_files,
        start=op.start, 
        end=op.end, 
        step=op.step,
    )


if __name__ == "__main__":
    main()



"""
These functions are generally here to help with saving stats of simulation
data
"""

from collections import defaultdict


def nested_dict():
    """
    Create a nested dictionary, usually used to save statistics from a
    simulation
    """
    return defaultdict(nested_dict)



"""
Store orientational entropies here
"""

from collections import Counter

import numpy as np

import waterEntropy.analysis.HB_labels as HBLabels
from waterEntropy.utils.helpers import nested_dict


class WaterOrientCalculator:
    """Calculate orientational entropy of water molecules here"""

    GAS_CONSTANT = 8.314  # (J/ (K mol)

    def __init__(self):
        self.Sorient = 0

    def add_data(self, shell_label: list, shell_values: dict):
        """
        Calculate the orientational entropy for water molecules with a
        specific shell type/environment.

        :param shell_label: the list or tuple of the shell labels of a
            coordination shell
        :param shell_values: the dictionary containing information about what
            neighbour types are donated to and accepted from
        """
        degeneracy = self.get_shell_degeneracy(shell_label)
        pD_dict = self.get_donor_acceptor_probabilities(shell_values, "donates_to")
        pA_dict = self.get_donor_acceptor_probabilities(shell_values, "accepts_from")
        Nc_eff, pbias_ave = self.get_reduced_neighbours_biases(
            degeneracy, pD_dict, pA_dict
        )
        self.Sorient = self.get_orientation_S(Nc_eff, pbias_ave)

    def get_shell_degeneracy(self, shell_label: list):
        """
        For a given labelled shell, find the degeneracy, i.e. count of each
        unique label in the shell

        :param shell_label: the list or tuple of the shell labels of a
            coordination shell
        """
        return Counter(shell_label)

    def get_donor_acceptor_probabilities(self, HB_vals_dict: dict, hb_selection: str):
        """
        For a given labelled shell type, find the probability of accepting from
        or donating to a given neighbour type over all other donors or acceptors

        :param vals_dict: dictionary containing the HB donors or acceptors and
            how often they occur for a given neighbour in a shell e.g.
            {"donates_to": {"labelled_donators": 0,},
            "accepts_from": {"labelled_acceptors": 0,}}
        :param degeneracy: dictionary containing the counts for each neighbour
            type in a shell
        :param hb_selection: string for determining which HB type to analyse,
            options are either "donates_to" or "accepts_from"
        """
        pHB_dict = nested_dict()
        if hb_selection in HB_vals_dict:
            total_hb_count = 0
            for c in HB_vals_dict[hb_selection].values():
                total_hb_count += c
            for hb_neighbour, count in sorted(list(HB_vals_dict[hb_selection].items())):
                pHB = self.get_hb_probability(count, total_hb_count)
                pHB_dict[hb_neighbour] = [pHB, count]
        return pHB_dict

    def get_reduced_neighbours_biases(
        self, degeneracy: dict, pD_dict: dict, pA_dict: dict
    ):
        # pylint: disable=too-many-locals
        r"""
        For a given labelled shell and the dictionary containing the counts for
        each neighbour type in the shell, the dictionaries for the acceptor and
        donator probabilities, use these to find the probability of accepting from
        or donating to over all other HBing to that particular neighbour type i:

        .. math::
            ppD_i = \frac{pD_i}{(pD_i + pA_i)}

        :param degeneracy: dictionary of shell constituent and count
        :param pD_dict: dictionary of neighbours donated to and how often that occurs
        :param pA_dict: dictionary of neighbours accepted from and how often that
            occurs
        """
        Nc_eff = 0
        pbiases = []
        # 1. iterate through each neighbour and the counts for how often they occur
        # in a shell
        for i, N_i in degeneracy.items():
            # 2. find if neighbour i is donated to or accepted from
            d = pD_dict[i] or [0, 0]
            a = pA_dict[i] or [0, 0]
            # 3. the donor/acceptor probabilities for a given neighbour i
            pD_i = d[0]
            pA_i = a[0]
            # 4. if the neighbour type i has both been donated to and accepted from
            # then work out probabilities of accepting from vs donating to
            if pD_i != 0 and pA_i != 0:
                sum_p = pD_i + pA_i
                # 5. work out the probabilities to donate to and accept from
                # over the sum of probabilities to donate and accept
                ppD_i = pD_i / sum_p
                ppA_i = pA_i / sum_p
                # 6. work out the effective number of available neighbour type i
                # where 0.25 is no bias in donating to/accepting from a given
                # neighbour
                N_i_eff = ppD_i * ppA_i / 0.25 * N_i
                # 7. sum the effective neighbours for all neighbour types i
                Nc_eff += N_i_eff
                # 8. work out the bias of donating vs accepting from neighbours of
                # type i, if both are equally likely, pbias = 0.25, if one is
                # more likely than the other the pbais < 0.25
                pbias = ppD_i * ppA_i
                # 9. append pbiases for each occurance of neighbour type i to a
                # list to get the averages later
                # for x in range(0, N_i):
                #     pbiases.append(pbias)
                pbiases.extend([pbias] * N_i)
            else:
                # 10. if a neighbour type i is not both donated to AND accepted from,
                # then it is not a neighbour involved in orientational
                # motion of the central UA and instead is either an anchor
                # or not involved in hydrogen bonding.
                # for x in range(0, N_i):
                #     pbiases.append(0)
                pbiases.extend([0] * N_i)

        # 11. find the average HB bias over all neighbours in the shell, where
        # zeros count towards the average
        pbias_ave = 0
        if len(pbiases) != 0:
            pbias_ave = sum(pbiases) / len(pbiases)

        return Nc_eff, pbias_ave

    def get_orientation_S(self, Nc_eff: float, pbias_ave: float):
        r"""
        Get the orientational entropy of water molecules, or any single UA molecule
        containing two hydrogen bond donors.

        .. math::
            S_\mathrm{orient} = \ln \Bigg(N_{c_\mathrm{eff}} ^ {(3 / 2)} \times \pi ^ {0.5}
            \times \frac{p_\mathrm{bias\_ave}}{2} \Bigg)

        This equation is modified from the previous theory of water molecule
        orientational entropy, where hydrogen bonding is accounted for by reducing
        the number of available neighbours.
        Here, the coordination shell neighbours available to hydrogen bond with
        are reduced to :math:`N_{c_\mathrm{eff}}`. The reduction in available
        HBing neighbours is
        calculated from statistics gathered from simulation trajectories, where
        neighbour types that are donated to or accepted from are counted.

        .. math::
            N_{c_\mathrm{eff}} = \sum_i \bigg((ppD_i \times ppA_i) \times \frac{N_i}{0.25} \bigg)

        where :math:`ppD_i` is the probability to donate to a given neighbour
        type :math:`i` compared to accepting from the same neighbour type.
        :math:`ppA_i` is the equivalent probability for accepting from neighbour
        type :math:`i`.

        :math:`p_\mathrm{bias\_ave}` is the average bias in accepting from and donating to
        a given neighbour type :math:`i`. If it is equally likely to donate to and
        accept from all neighbour types, then :math:`p_\mathrm{bias\_ave}=0.25`, but if there is
        any bias in preferentially donating to, accepting from or not HBing to a
        neighbour at all, then :math:`p_\mathrm{bias\_ave} < 0.25`.

        .. math::
            p_\mathrm{bias\_ave} = \frac{\sum_i(ppD_i * ppA_i)}{N_c}

        Both :math:`N_{c_\mathrm{eff}}` and :math:`p_\mathrm{bias\_ave}` are
        used to reduce the orientational entropy of the central molecule by
        accounting for the HBing observed in a simulation.

        :param Nc_eff: effective number of available neighbours to rotate around
        :param pbias_ave: the average biasing in hydrogen bonding donating or
            accepting over all shell neighbours
        """
        S_orient = 0
        if Nc_eff != 0:
            S_orient = np.log((Nc_eff) ** (3 / 2) * np.pi**0.5 * pbias_ave / 2)
        S_orient = max(S_orient, 0)
        return S_orient * self.GAS_CONSTANT

    def get_hb_probability(self, count: int, total_count: int):
        r"""
        For a given HB donor/acceptor, find the probablity of that HB occurring
        in a given unique labelled shell over all the other acceptors or donors,
        so

        .. math::
            pA_i = \frac{N_{i_A}}{\sum_i(N_{i_A})}

        or

        .. math::
            pD_i = \frac{N_{i_D}}{\sum_i(N_{i_D})}

        :param count: number of HB donors/acceptors for a given neighbour type
        :param total_count: total number of either acceptor or donor counts over
            all neighbour types
        """
        p_i = count / total_count
        return p_i


class Orientations:
    """
    Store orientational entropies here
    """

    def __init__(self):
        self.resid_labelled_Sorient = nested_dict()
        self.resname_labelled_Sorient = nested_dict()

    def add_data(self, hb_labels: HBLabels):
        """Add orientational entropy data"""
        self.get_orientational_entropy_from_dict(
            hb_labels.labelled_shell_counts, self.resname_labelled_Sorient
        )
        self.get_resid_orientational_entropy_from_dict(
            hb_labels.resid_labelled_shell_counts
        )

    def get_resid_orientational_entropy_from_dict(self, resid_labelled_dict: dict):
        """
        For a given dictionary containing labelled shells and HBing within the
        shell with format:

        resid_labelled_dict = {"nearest_resid": {"resname":
                {("labelled_shell"): {"shell_count": 0,
                "donates_to": {"labelled_donators": 0,},
                "accepts_from": {"labelled_acceptors": 0,}
                }}}

        Get the orientational entropy of the molecules in this dict

        :param resid_labelled_dict: dictionary of format dict2 containing labelled
            coordination shells and HB donating and accepting
        """
        for resid, shell_label_key in sorted(list(resid_labelled_dict.items())):
            self.get_orientational_entropy_from_dict(
                shell_label_key, self.resid_labelled_Sorient[resid]
            )

    def get_running_average(
        self, value: float, count: int, running_average_value: float, count_stored: int
    ):
        """
        For a given value, get it's running average from the current value

        :param value: the value that needs to be added to the running average
        :param count: the number of times the value occurs from statistics
        :param running_average_value: the currently stored running average
        :param count_stored: the currently stored count for the running average
        """
        new_count_stored = count_stored + count
        new_running_average = (
            value * count + running_average_value * count_stored
        ) / new_count_stored

        return new_running_average, new_count_stored

    def get_orientational_entropy_from_dict(
        self, labelled_dict: dict, Sorient_dict: dict
    ):
        # pylint: disable=too-many-locals
        """
        For a given dictionary containing labelled shells and HBing within the
        shell with format:

        labelled_dict = {"resname": {("labelled_shell"): {"shell_count": 0,
                                    "donates_to": {"labelled_donators": 0,},
                                    "accepts_from": {"labelled_acceptors": 0,}
                                    }}}

        Get the orientational entropy of the molecules in this dict

        :param labelled_dict: dictionary of format dict1 containing labelled
            coordination shells and HB donating and accepting
        """
        for resname, shell_label_key in sorted(list(labelled_dict.items())):
            Sorient_ave, tot_count = 0, 0
            for shell_label, values in sorted(list(shell_label_key.items())):
                water = WaterOrientCalculator()
                water.add_data(shell_label, values)
                Sorient_ave, tot_count = self.get_running_average(
                    water.Sorient, values["shell_count"], Sorient_ave, tot_count
                )
            Sorient_dict[resname] = [Sorient_ave, tot_count]


def print_Sorient_dicts(Sorient_dict: dict):
    """
    Print the orientational entropies of interfacial solvent

    :param Sorient_dict: dictionary containing orientational entropy values
    """
    for resid, resname_key in sorted(list(Sorient_dict.items())):
        for resname, [Sor, count] in sorted(list(resname_key.items())):
            print(resid, resname, Sor, count)



"""
Get the rotated mass-weighted forces and inertia-weighted torques from
transformed atom positons and forces and save these as covariance matrices
in a Covariances class instance.
"""

import numpy as np

from waterEntropy.entropy.convariances import CovarianceCollection
import waterEntropy.maths.transformations as Transformation
import waterEntropy.utils.selections as Select


def get_forces_torques(
    covariances: CovarianceCollection, molecule, nearest: str, system
):
    # pylint: disable=too-many-locals
    """
    Calculate the covariance matrices of molecules and populate these
    in the covariances instance. Molecules are grouped based on the molecule
    name and the name of the molecule that is nearest to it.

    :param covariances: instance of CovarianceCollection class
    :param molecule: MDAnalysis instance of a molecule
    :param nearest: name of nearest molecule
    :param system: MDAnalysis instance of whole system
    """
    # 1. get the length scale of the molecule
    molecule_scale = Select.guess_length_scale(molecule)
    # 1b. get the value to scale the forces and torque matrices with
    if molecule_scale == "single_UA":
        scale_covariance = 0.5
    else:
        # don't scale larger molecules automatically as there may be longer
        # lengthscales in the hierarchy
        scale_covariance = 1
    # 2. Get the axes of the molecule
    principal_axes, MOI_axis = Transformation.get_axes(molecule, molecule_scale)
    # 3. Get the center point of the molecule
    center_of_mass = molecule.center_of_mass()
    # 4. calculate the torque from the forces and axes
    torque = Transformation.get_torques(
        molecule, center_of_mass, principal_axes, MOI_axis
    )
    # 5. calculate the mass weighted forces
    mass_weighted_force = Transformation.get_mass_weighted_forces(
        molecule, principal_axes
    )
    # 6. calculate the covariance matrices
    F_cov_matrix = Transformation.get_covariance_matrix(
        mass_weighted_force, scale_covariance
    )
    T_cov_matrix = Transformation.get_covariance_matrix(torque, scale_covariance)
    # add the covariances to the class instance
    covariances.add_data(nearest, molecule.resnames[0], F_cov_matrix, T_cov_matrix)

    # not applicable to water
    if molecule_scale == "multiple_UAs":
        UAs = Select.find_molecule_UAs(molecule)
        F_cov_matrices, T_cov_matrices = [], []
        for UA in UAs:
            # find the axes based on what the UA is bonded to
            custom_axes, position_vector = Transformation.get_bonded_axes(
                system, UA, system.dimensions[:3]
            )
            if custom_axes is not None:  # ignore if UA is only bonded to one other UA
                # set the center of mass as the coordinates of the UA
                center_of_mass = UA.position
                # calcuate the torques using the custom axes based on bonds
                torque = Transformation.get_torques(
                    molecule, center_of_mass, custom_axes, position_vector
                )
                # calcuate the mass weighted forces using the custom axes based on bonds
                mass_weighted_force = Transformation.get_mass_weighted_forces(
                    molecule, custom_axes
                )
                # calculate and append the covariances matrices
                F_cov_matrix = Transformation.get_covariance_matrix(
                    mass_weighted_force, scale_covariance
                )
                T_cov_matrix = Transformation.get_covariance_matrix(
                    torque, scale_covariance
                )
                F_cov_matrices.append([F_cov_matrix])
                T_cov_matrices.append([T_cov_matrix])
        # populate the covariances to the class instance
        # TO-DO: set a class instance specifically for the UA length scale of molecules > 1UA
        covariances.add_data(
            molecule,
            molecule,
            np.concatenate(F_cov_matrices, axis=0),
            np.concatenate(T_cov_matrices, axis=0),
        )



"""
These functions calculate orientational entropy from labelled
coordination shells
"""

import waterEntropy.analysis.HB as HBond
import waterEntropy.analysis.HB_labels as HBLabels
import waterEntropy.analysis.RAD as RADShell
import waterEntropy.analysis.shell_labels as RADLabels
from waterEntropy.analysis.shells import ShellCollection
from waterEntropy.entropy.convariances import CovarianceCollection
from waterEntropy.entropy.orientations import Orientations
from waterEntropy.entropy.vibrations import Vibrations
from waterEntropy.recipes.forces_torques import get_forces_torques
from waterEntropy.utils.helpers import nested_dict
import waterEntropy.utils.selections as Selections


def find_interfacial_solvent(solutes, system, shells: ShellCollection):
    """
    For a given set of solute molecules, find the RAD shells for each UA in the
    molecules, if a solvent molecule is in the RAD shell, then save the solvent
    atom index to a list. A solvent is defined as molecule that constitutes a
    single UA. These solvent molecule are defined as interfacial molecules.

    :param solutes: mdanalysis instance of a selection of atoms in solute
        molecules that are greater than one UA
    :param system: mdanalysis instance of atoms in a frame
    :param shells: ShellCollection instance
    """
    solvent_indices = []
    molecules = solutes.fragments  # fragments is mdanalysis equiv to molecules
    for molecule in molecules:
        # 1. find heavy atoms in the molecule
        UAs = Selections.find_molecule_UAs(molecule)
        for atom in UAs:
            # 2. find the shell of each UA atom in a molecule
            shell = RADShell.get_RAD_shell(
                atom, system, shells
            )  # get the molecule UA shell
            shell_indices = shell.UA_shell
            # 3. for each neighbour in the RAD shell, find single UA molecules
            shell_atoms = Selections.get_selection(system, "index", shell_indices)
            for neighbour_atom in shell_atoms:
                # create an atom group to find molecule/fragment
                neighbour_atomGroup = Selections.get_selection(
                    system, "index", [neighbour_atom.index]
                )
                neighbour_molecule = neighbour_atomGroup.fragments[0]
                neighbour_UAs = Selections.find_molecule_UAs(
                    neighbour_molecule.fragments[0]
                )  # 1 length list
                if len(neighbour_UAs) == 1:  # single UA molecule is a solvent
                    # 5. add single UA molecules into the solvent indices list
                    if neighbour_atom.index not in solvent_indices:
                        solvent_indices.append(neighbour_atom.index)
                    else:
                        continue
                else:
                    continue
    return solvent_indices


def get_interfacial_water_orient_entropy(system, start: int, end: int, step: int):
    # pylint: disable=too-many-locals
    """
    For a given system, containing the topology and coordinates of molecules,
    find the interfacial water molecules around solutes and calculate their
    orientational entropy if there is a solute atom in the solvent coordination
    shell.

    :param system: mdanalysis instance of atoms in a frame
    :param start: starting frame number
    :param end: end frame number
    :param step: steps between frames
    """
    # don't need to include the frame_solvent_indices dictionary
    frame_solvent_indices = nested_dict()
    # initialise the Covariance class instance to store covariance matrices
    covariances = CovarianceCollection()
    # initialise the Vibrations class instance to store vibrational entropies
    vibrations = Vibrations(temperature=298, force_units="kcal")
    hb_labels = HBLabels.HBLabelCollection()
    # pylint: disable=unused-variable
    for ts in system.trajectory[start:end:step]:
        # initialise the RAD and HB class instances to store shell information
        shells = ShellCollection()
        HBs = HBond.HBCollection()
        # 1. find > 1 UA molecules in system, these are the solutes
        resid_list = Selections.find_solute_molecules(system)
        solutes = Selections.get_selection(system, "resid", resid_list)
        # 2. find the interfacial solvent molecules that are 1 UA in size
        #   and are in the RAD shell of any solute
        solvent_indices = find_interfacial_solvent(solutes, system, shells)
        first_shell_solvent = Selections.get_selection(system, "index", solvent_indices)
        # 3. iterate through first shell solvent and find their RAD shells,
        #   HBing in the shells and shell labels
        for solvent in first_shell_solvent:
            # print(solvent)
            # 3a. find RAD shell of interfacial solvent
            shell = RADShell.get_RAD_shell(solvent, system, shells)
            # 3b. find HBing in the shell
            HBond.get_shell_HBs(shell, system, HBs, shells)
            # 3c. find RAD shell labels
            shell = RADLabels.get_shell_labels(solvent.index, system, shell, shells)
            # 3d. find HB labels
            HBLabels.get_HB_labels(solvent.index, system, HBs, shells)
            if shell.nearest_nonlike_idx is not None:
                # 3e. populate the labels into a dictionary for stats
                # only if a different atom is in the RAD shell
                nearest = system.atoms[shell.nearest_nonlike_idx]
                nearest_resid = nearest.resid
                nearest_resname = nearest.resname
                hb_labels.add_data(
                    nearest_resid,
                    nearest_resname,
                    shell.labels,
                    shell.donates_to_labels,
                    shell.accepts_from_labels,
                )
                frame_solvent_indices = save_solvent_indices(
                    ts.frame,
                    shell.atom_idx,
                    nearest_resid,
                    nearest_resname,
                    frame_solvent_indices,
                )
                # 3f. calculate the running average of force and torque
                # covariance matrices
                solvent_molecule = system.atoms[solvent.index].fragment  # get molecule
                get_forces_torques(
                    covariances,
                    solvent_molecule,
                    f"{nearest_resname}_{nearest_resid}",
                    system,
                )

    # 4. get the orientational entropy of interfacial waters and save
    #   them to a dictionary
    # TO-DO: add average Nc in Sorient dict
    # Sorient_dict = Orient.get_resid_orientational_entropy_from_dict(
    #     hb_labels.resid_labelled_shell_counts
    # )
    Sorients = Orientations()
    Sorients.add_data(hb_labels)
    Sorient_dict = Sorients.resid_labelled_Sorient
    # 5. Get the vibrational entropy of interfacial waters
    vibrations.add_data(covariances, diagonalise=True)

    return (
        Sorient_dict,
        covariances,
        vibrations,
        frame_solvent_indices,
    )


def save_solvent_indices(
    frame: int,
    atom_idx: int,
    nearest_resid: int,
    nearest_resname: str,
    frame_solvent_indices: dict,
):
    """
    Save the solvent indices at interfaces per frame into a dictionary

    :param frame: frame number of analysed frame
    :param atom_idx: solvent atom index
    :param nearest_resid: residue of number of nearest solute molecule
    :param nearest_resname: residue name of nearest solute molecule
    :param frame_solvent_indices: the dictionary to populate
    """
    if nearest_resid not in frame_solvent_indices[frame][nearest_resname]:
        frame_solvent_indices[frame][nearest_resname][nearest_resid] = []
    frame_solvent_indices[frame][nearest_resname][nearest_resid].append(atom_idx)
    return frame_solvent_indices


def print_frame_solvent_dicts(frame_solvent_indices: dict):
    """
    Print the interfacial solvent for each analysed frame

    :param frame_solvent_indices: dictionary containing solvent indices in the
        first shell of solute atoms over each frame analysed
    """
    for frame, resname_key in sorted(list(frame_solvent_indices.items())):
        for resname, resid_key in sorted(list(resname_key.items())):
            for resid, solvents in sorted(list(resid_key.items())):
                print(frame, resname, resid, len(solvents), solvents)



"""
Functions for common trigonometric calculations
"""

import MDAnalysis
import numpy as np

import waterEntropy.utils.selections as Selections


def get_neighbourlist(
    atom: np.ndarray, neighbours, dimensions: np.ndarray, max_cutoff=9e9
):
    """
    Use MDAnalysis to get distances between an atom and neighbours within
    a given cutoff. Each atom index pair sorted by distance are outputted.

    :param atom: (3,) array of an atom coordinates.
    :param neighbours: MDAnalysis array of heavy atoms in the system,
        not the atom itself and not bonded to the atom.
    :param dimensions: (6,) array of system box dimensions.
    :param max_cutoff: set the maximum cutoff value for finding neighbour distances
    """
    # check atom coords are not in neighbour coords list
    if not (atom == neighbours.positions).all(axis=1).any():
        pairs, distances = MDAnalysis.lib.distances.capped_distance(
            atom,
            neighbours.positions,
            max_cutoff=max_cutoff,
            min_cutoff=None,
            box=dimensions,
            method=None,
            return_distances=True,
        )
        neighbour_indices = neighbours[pairs[:][:, 1]].indices
        sorted_distances, sorted_indices = zip(
            *sorted(zip(distances, neighbour_indices), key=lambda x: x[0])
        )
        return np.array(sorted_indices), np.array(sorted_distances)
    raise ValueError(
        f"Atom coordinates {atom} in neighbour list {neighbours.positions[:10]}"
    )


def get_sorted_neighbours(i_idx: int, system, max_cutoff=25):
    """
    For a given atom, find neighbouring united atoms from closest to furthest
    within a given cutoff.

    :param i_idx: idx of atom i
    :param system: mdanalysis instance of atoms in a frame
    :param max_cutoff: set the maximum cutoff value for finding neighbours
    """
    i_coords = system.atoms.positions[i_idx]
    # 1. get the heavy atom neighbour distances within a given distance cutoff
    # CHECK Find out which of the options below is better for RAD shells
    #       Should the central atom bonded UAs be allowed to block?
    #       This was not done in original code, keep the same here
    neighbours = system.select_atoms(
        f"""mass 2 to 999 and not index {i_idx}
                                    and not bonded index {i_idx}"""
        # f"""mass 2 to 999 and not index {i_idx}"""  # bonded UAs can block
    )
    # 2. Get the neighbours sorted from closest to furthest
    sorted_indices, sorted_distances = get_neighbourlist(
        i_coords, neighbours.atoms, system.dimensions, max_cutoff
    )
    return sorted_indices, sorted_distances


def get_shell_neighbour_selection(
    shell, donator, system, heavy_atoms=True, max_cutoff=10
):
    """
    get shell neighbours ordered by ascending distance, this is used for
    finding possible hydrogen bonding neighbours.

    :param shell: the instance for class waterEntropy.neighbours.RAD.RAD
        containing coordination shell neighbours
    :param donator: the mdanalysis object for the donator
    :param system: mdanalysis instance of all atoms in current frame
    :param heavy_atoms: consider heavy atoms in a shell as neighbours
    :max_cutoff: set the maximum cutoff value for finding neighbours
    """
    # 1a. Select heavy atoms in shell, can only donate to heavy atoms in the shell
    neighbours = Selections.get_selection(system, "index", shell.UA_shell)
    if not heavy_atoms:
        # 1b. Select all atoms in the shell, included bonded to atoms (Hs included)
        #   Can donate to any atoms in a shell
        all_shell_bonded = neighbours[:].bonds.indices
        all_shell_indices = list(set().union(*all_shell_bonded))
        # can donate to any atom in the shell
        neighbours = Selections.get_selection(system, "index", all_shell_indices)
    # 1c. can donate to any neighbours outside a shell, not used
    # neighbours = system.select_atoms(f"all and not index {donator.index} and not bonded index {donator.index}")
    # 2. Get the neighbours sorted from closest to furthest
    sorted_indices, sorted_distances = get_neighbourlist(
        donator.position, neighbours, system.dimensions, max_cutoff
    )
    return sorted_indices, sorted_distances


def get_angle(a: np.ndarray, b: np.ndarray, c: np.ndarray, dimensions: np.ndarray):
    """
    Get the angle between three atoms, taking into account PBC.

    :param a: (3,) array of atom cooordinates
    :param b: (3,) array of atom cooordinates
    :param c: (3,) array of atom cooordinates
    :param dimensions: (3,) array of system box dimensions.
    """
    ba = np.abs(a - b)
    bc = np.abs(c - b)
    ac = np.abs(c - a)
    ba = np.where(ba > 0.5 * dimensions, ba - dimensions, ba)
    bc = np.where(bc > 0.5 * dimensions, bc - dimensions, bc)
    ac = np.where(ac > 0.5 * dimensions, ac - dimensions, ac)
    dist_ba = np.sqrt((ba**2).sum(axis=-1))
    dist_bc = np.sqrt((bc**2).sum(axis=-1))
    dist_ac = np.sqrt((ac**2).sum(axis=-1))
    cosine_angle = (dist_ac**2 - dist_bc**2 - dist_ba**2) / (-2 * dist_bc * dist_ba)
    return cosine_angle


def get_distance(a: np.ndarray, b: np.ndarray, dimensions: np.ndarray):
    """
    Calculates distance and accounts for PBCs.

    :param a: (3,) array of atom cooordinates
    :param b: (3,) array of atom cooordinates
    :param dimensions: (3,) array of system box dimensions.
    """
    delta = np.abs(b - a)
    delta = np.where(delta > 0.5 * dimensions, delta - dimensions, delta)
    distance = np.sqrt((delta**2).sum(axis=-1))
    return distance


def get_vector(a: np.ndarray, b: np.ndarray, dimensions: np.ndarray):
    """
    get vector of two coordinates over PBCs.

    :param a: (3,) array of atom cooordinates
    :param b: (3,) array of atom cooordinates
    :param dimensions: (3,) array of system box dimensions.
    """
    delta = b - a
    delta_wrapped = []
    for delt, box in zip(delta, dimensions):
        if delt < 0 and delt < 0.5 * box:
            delt = delt + box
        if delt > 0 and delt > 0.5 * box:
            delt = delt - box
        delta_wrapped.append(delt)
    delta_wrapped = np.array(delta_wrapped)

    return delta_wrapped



"""
Store vibrational entropy from covariance matrices
"""

import numpy as np
from numpy import linalg as LA

from waterEntropy.entropy.convariances import CovarianceCollection
from waterEntropy.utils.helpers import nested_dict


class Vibrations:
    r"""
    Store molecule vibrational entropy information here, where

    .. math::
        S^{\mathrm{vib}} = S^{\mathrm{trans}} + S^{\mathrm{rot}}

    Entropies are calculated from the covariance matrices inputted in the
    add_data method.
    """

    # constants
    BOLTZMANN = 1.38064852e-23  # (J / K)
    HBAR = 1.0545718e-34  # (J s)
    AVOGADRO = 6.022e23
    GAS_CONSTANT = 8.314  # (J/ (K mol)
    # conversions
    CALORIE_TO_JOULE = 4184
    KG_TO_AMU = 6.02214086e26
    PER_MOLE_TO_PER_MOLECULE = 6.02214086e23
    ANGSTROM_TO_METRE = 1e-10
    KJ_TO_J = 1000
    NANOMETRE_TO_METRE = 1e-9

    def __init__(self, temperature, force_units):
        self.temperature = temperature
        self.force_units = force_units
        self.translational_freq = nested_dict()
        self.rotational_freq = nested_dict()
        self.translational_S = nested_dict()
        self.rotational_S = nested_dict()

    def add_data(self, covariances: CovarianceCollection, diagonalise=True):
        """
        Calculate and add the frequencies and vibrational entropies to class
        instance

        :param covariances: instance of CovarianceCollection class
        :param diagonalise: diagonalise the covariance matrix rather than getting
            the eigenvalues and eigenvectors
        """
        for (nearest, molecule_name), force_covariance in covariances.forces.items():
            torque_covariance = covariances.torques[(nearest, molecule_name)]
            force_Svib, force_frequency = Vibrations.get_data(
                self, force_covariance, diagonalise
            )
            torque_Svib, torque_frequency = Vibrations.get_data(
                self, torque_covariance, diagonalise
            )
            Vibrations.populate_dicts(
                self, nearest, molecule_name, force_Svib, self.translational_S
            )
            Vibrations.populate_dicts(
                self, nearest, molecule_name, torque_Svib, self.rotational_S
            )
            Vibrations.populate_dicts(
                self, nearest, molecule_name, force_frequency, self.translational_freq
            )
            Vibrations.populate_dicts(
                self, nearest, molecule_name, torque_frequency, self.rotational_freq
            )

    def get_data(self, covariance: CovarianceCollection, diagonalise: bool):
        """
        Get the frequencies and vibrational entropies from the covariance matrix

        :param covariances: instance of CovarianceCollection class
        :param diagonalise: whether to diagonalise the covariance matrix
        """
        # convert the covariance matrix to the correct units
        if self.force_units == "gromacs":
            covariance *= self.kJ_conversion()
        else:
            covariance *= self.kcal_conversion()
        Svib, frequency = Vibrations.calculate_entropies(self, covariance, diagonalise)

        return Svib, frequency

    def populate_dicts(self, nearest, molecule_name: str, variable, variable_dict):
        """
        Add frequencies / entropies to class instance dictionaries

        :param nearest: the value for grouping molecules into a class. e.g. the
            name of the nearest molecule
        :param molecule_name: the name of the molecule being populated into the
            dictionaries
        :param variable: the variable being added to the variable dictionary
        :param variable_dict: the dictionary being updated with a variable
        """
        if (nearest, molecule_name) not in variable_dict:
            variable_dict[(nearest, molecule_name)] = variable
        else:
            np.append(variable_dict[(nearest, molecule_name)], variable)

    def kJ_conversion(self):
        """
        Convert force covariance matrix from kJ/(mol nm m^2) to J/mol
        SI: kg.m/s^2 = N
        The forces are squared, so do the same for the constant
        """
        numerator = (self.KJ_TO_J**2) * self.KG_TO_AMU
        denominator = (self.AVOGADRO**2) * (self.NANOMETRE_TO_METRE**2)
        conversion = numerator / denominator

        return conversion

    def kcal_conversion(self):
        """
        Convert force covariance matrix from Kcal/(mol Ang m^2) to J/mol
        SI: kg.m/s^2 = N
        The forces are squared, so do the same for the constant
        """
        numerator = (self.CALORIE_TO_JOULE**2) * self.KG_TO_AMU
        denominator = (self.AVOGADRO**2) * (self.ANGSTROM_TO_METRE**2)
        conversion = numerator / denominator

        return conversion

    def calculate_entropies(self, covariance: CovarianceCollection, diagonalise: bool):
        r"""
        Calculate the vibrational entropies from the equation of a quantum
        harmonic oscillator:

        .. math::
            S^{\mathrm{vib}} = k_{\mathrm{B}} \sum_{i=1}^{N_\mathrm{vib}} \\
            \Bigg( \frac{h\nu_i/k_{\mathrm{B}}T}{e^{h\nu_i/k_{\mathrm{B}}T}-1} \\
            - \ln(1 - e^{-h\nu_i/k_{\mathrm{B}}T} ) \Bigg)

        where:

        - :math:`k_{\mathrm{B}}` is the Boltzmann constant :math:`\mathrm{J/K}` (Joule per Kelvin),

        :param covariances: instance of CovarianceCollection class
        :param diagonalise: whether to diagonalise the covariance matrix
        """
        frequency, eigenvalues = Vibrations.calculate_frequencies(
            self, covariance, diagonalise
        )
        a = (self.HBAR * frequency) / (self.BOLTZMANN * self.temperature)
        b = np.where(a != 0, np.exp(a) - 1, 0.0)
        c = np.where(a != 0, np.log(1 - np.exp(-a)), 0.0)
        Svib = np.where((b != 0) & (c != 0), a / b - c, 0)

        return Svib * self.GAS_CONSTANT, eigenvalues

    def calculate_frequencies(
        self, covariance: CovarianceCollection, diagonalise: bool
    ):
        """
        Calculate the frequencies for each molecule from the force and torque
        covariance matrices.

        :param covariances: instance of CovarianceCollection class
        :param diagonalise: whether to diagonalise the covariance matrix
        """
        if diagonalise:
            eigenvalues = covariance.diagonal()  # get matrix diagonals only
        else:
            eigenvalues, _eigenvectors = LA.eig(covariance)
        # Replace eigenvalues numbers with 0
        filtered_eigenvalues = np.where(np.isreal(eigenvalues), eigenvalues.real, 0)
        # Replace negative eigenvalues with 0
        filtered_eigenvalues = np.where(
            filtered_eigenvalues >= 0, filtered_eigenvalues, 0
        )
        # calculate frequencies from filtered eigenvalues
        frequency = (filtered_eigenvalues / (self.BOLTZMANN * self.temperature)) ** 0.5

        return frequency, eigenvalues


def print_Svib_data(vibrations: Vibrations, covariances: CovarianceCollection):
    """
    Print the orientational entropies of interfacial solvent

    :param vibrations: instance of Vibrations class
    :param covariances: instance of CovarianceCollection class
    """
    for near_solvent_name, Strans in vibrations.translational_S.items():
        near = near_solvent_name[0]
        solvent_name = near_solvent_name[1]
        # forces = covariances.forces[near_solvent_name]
        # torques = covariances.torques[near_solvent_name]
        Strans = vibrations.translational_S[near_solvent_name]
        Srot = vibrations.rotational_S[near_solvent_name]
        # trans_freqs = vibrations.translational_freq[near_solvent_name]
        # rot_freqs = vibrations.rotational_freq[near_solvent_name]
        counts = covariances.counts[near_solvent_name]
        print(near, solvent_name, sum(Strans), sum(Srot), counts)



"""
Functions for common MDAnalysis selections
"""

import numpy as np


def get_selection(system, selection_type: str, indices: list[int]):
    """
    For a list of indices, turn this list into a string and use it
    to select by a particular MDAnalysis selection type e.g. resid or index

    :param system: mdanalysis instance of atoms in a frame
    :param selection_type: string for what selection to make in mdanalysis, e.g
        index, resid
    :param indices: list of numbers corresponding to the atoms or resids to
        select
    """
    idx_str = " ".join(str(n) for n in indices)
    selection = system.select_atoms(f"""{selection_type} {idx_str}""")
    return selection


def find_molecule_UAs(molecule):
    """
    For a given molecule, return the heavy atoms it constitutes

    :param molecule: mdanalysis instance of atoms in a frame
    """
    UAs = molecule.select_atoms("mass 2 to 999")
    return UAs


def find_solute_molecules(system):
    """
    For a given system, find molecules containing more than one UA or is not
    a single UA molecule that contains an oxygen atom and return
    the resids for these molecules.

    :param system: mdanalysis instance of atoms in a frame
    """
    atom = system.select_atoms("all")
    molecules = atom.fragments
    solute_molecule_resid_list = []
    for molecule in molecules:
        for res in molecule.residues:
            UAs = atom.select_atoms(f"resid {res.resid} and mass 2 to 999")
            if len(UAs) > 1:
                solute_molecule_resid_list.append(res.resid)
            # if heavy atom is not oxygen, treat as solute
            if len(UAs) == 1 and np.floor(UAs[0].mass) != 16:
                solute_molecule_resid_list.append(res.resid)

    return solute_molecule_resid_list


def find_bonded_heavy_atom(atom_idx: int, system):
    """
    for a given atom, if it is a hydrogen, find what heavy atom it is bonded to

    :param atom_idx: atom index to find bonded heavy atom for
    :param system: mdanalysis instance of all atoms in current frame
    """
    atom = system.atoms[atom_idx]
    if atom.mass < 1.1:
        bonded_atoms = system.select_atoms(f"bonded index {atom_idx}")
        bonded_heavy_atoms = bonded_atoms.select_atoms("mass 2 to 999")
        bonded_heavy_atom = bonded_heavy_atoms[0]  # should be a list of one
    else:
        bonded_heavy_atom = atom
    return bonded_heavy_atom


def find_bonded_atoms(atom_idx: int, system):
    """
    for a given atom, find its bonded heavy and H atoms

    :param atom_idx: atom index to find bonded heavy atom for
    :param system: mdanalysis instance of all atoms in current frame
    """
    bonded_atoms = system.select_atoms(f"bonded index {atom_idx}")
    bonded_heavy_atoms = bonded_atoms.select_atoms("mass 2 to 999")
    bonded_H_atoms = bonded_atoms.select_atoms("mass 1 to 1.1")
    return bonded_heavy_atoms, bonded_H_atoms


def guess_length_scale(molecule):
    """Guess what the length scale of the molecule is

    :param molecule: MDAnalysis instance of molecule
    """
    molecule_scale = None
    UAs = find_molecule_UAs(molecule)
    if len(UAs) == 1:
        molecule_scale = "single_UA"
    elif len(UAs) > 1:
        if len(molecule.atoms.residues) > 1:
            molecule_scale = "polymer"
        else:
            molecule_scale = "multiple_UAs"
    else:
        molecule_scale = "no_UA"
    return molecule_scale



"""
Store the labelled neighbours that are donated to and accepted from the central
atom in a shell. Shells with the same neighbours are grouped together and
along with the counts for donating to and accepting from each neighbour type.
These labelled HB neighbours are used to calculate orientational
entropy of water molecules.
"""

from waterEntropy.analysis.HB import HBCollection
from waterEntropy.analysis.shells import ShellCollection
from waterEntropy.utils.helpers import nested_dict
import waterEntropy.utils.selections as Selections


class HBLabelCollection:
    """
    Labelled shell counts used for Sorient.
    The counts are placed into two dictionaries used for statistics later,
    the structure of these two dictionaries are as follows:

    dict1 = {"resname": {("labelled_shell"): {"shell_count": 0,
                                "donates_to": {"labelled_donators": 0,},
                                "accepts_from": {"labelled_acceptors": 0,}
                                }}}
    dict2 = {"nearest_resid": {"resname":
                                {("labelled_shell"): {"shell_count": 0,
                                "donates_to": {"labelled_donators": 0,},
                                "accepts_from": {"labelled_acceptors": 0,}
                                }}}
    """

    def __init__(self):
        self.labelled_shell_counts = nested_dict()  # save shell instances in here
        self.resid_labelled_shell_counts = nested_dict()  # save shell instances in here

    def add_data(self, resid, resname, labelled_shell, donates_to, accepts_from):
        # pylint: disable=too-many-arguments
        """Add data to class dictionaries"""
        self.add_shell_counts(resid, resname, labelled_shell)
        self.add_donates_to(resid, resname, labelled_shell, donates_to)
        self.add_accepts_from(resid, resname, labelled_shell, accepts_from)

    def add_shell_counts(self, resid, resname, labelled_shell):
        """
        Add a labelled shell to a dictionary that keeps track of the counts
        for each labelled shell type with constituents alpha-numerically
        ordered

        :param self: class instance
        :param resid: residue id of nearest nonlike atom in the labelled shell
        :param resname: residue name of nearest nonlike atom in the labelled shell
        :param labelled_shell: coordination shell with labelled neighbours
        """
        labelled_shell = tuple(sorted(labelled_shell))
        if "shell_count" not in self.labelled_shell_counts[resname][labelled_shell]:
            self.labelled_shell_counts[resname][labelled_shell]["shell_count"] = 1
        else:
            self.labelled_shell_counts[resname][labelled_shell]["shell_count"] += 1

        if (
            "shell_count"
            not in self.resid_labelled_shell_counts[resid][resname][labelled_shell]
        ):
            self.resid_labelled_shell_counts[resid][resname][labelled_shell][
                "shell_count"
            ] = 1
        else:
            self.resid_labelled_shell_counts[resid][resname][labelled_shell][
                "shell_count"
            ] += 1

    def add_donates_to(self, resid, resname, labelled_shell, donates_to):
        """
        Add a labelled neighbours donated to in a dictionary that keeps
        track of the counts for each labelled shell type with
        constituents alpha-numerically ordered

        :param self: class instance
        :param resid: residue id of nearest nonlike atom in the labelled shell
        :param resname: residue name of nearest nonlike atom in the labelled shell
        :param labelled_shell: coordination shell with labelled neighbours
        :param donates_to: list of labelled neighbours that are donated to
        """
        labelled_shell = tuple(sorted(labelled_shell))
        # donates_to = tuple(sorted(donates_to))
        for a in donates_to:
            if (
                a
                not in self.labelled_shell_counts[resname][labelled_shell]["donates_to"]
            ):
                self.labelled_shell_counts[resname][labelled_shell]["donates_to"][a] = 1
            else:
                self.labelled_shell_counts[resname][labelled_shell]["donates_to"][
                    a
                ] += 1

            if (
                a
                not in self.resid_labelled_shell_counts[resid][resname][labelled_shell][
                    "donates_to"
                ]
            ):
                self.resid_labelled_shell_counts[resid][resname][labelled_shell][
                    "donates_to"
                ][a] = 1
            else:
                self.resid_labelled_shell_counts[resid][resname][labelled_shell][
                    "donates_to"
                ][a] += 1

    def add_accepts_from(self, resid, resname, labelled_shell, accepts_from):
        """
        Add a labelled neighbours accepted from to in a dictionary that keeps
        track of the counts for each labelled shell type with
        constituents alpha-numerically ordered

        :param self: class instance
        :param resid: residue id of nearest nonlike atom in the labelled shell
        :param resname: residue name of nearest nonlike atom in the labelled shell
        :param labelled_shell: coordination shell with labelled neighbours
        :param accepts_from: list of labelled neighbours that are accepted_from
        """
        labelled_shell = tuple(sorted(labelled_shell))
        for d in accepts_from:
            if (
                d
                not in self.labelled_shell_counts[resname][labelled_shell][
                    "accepts_from"
                ]
            ):
                self.labelled_shell_counts[resname][labelled_shell]["accepts_from"][
                    d
                ] = 1
            else:
                self.labelled_shell_counts[resname][labelled_shell]["accepts_from"][
                    d
                ] += 1

            if (
                d
                not in self.resid_labelled_shell_counts[resid][resname][labelled_shell][
                    "accepts_from"
                ]
            ):
                self.resid_labelled_shell_counts[resid][resname][labelled_shell][
                    "accepts_from"
                ][d] = 1
            else:
                self.resid_labelled_shell_counts[resid][resname][labelled_shell][
                    "accepts_from"
                ][d] += 1


def get_HB_labels(atom_idx: int, system, HBs: HBCollection, shells: ShellCollection):
    """
    For a given central atom, get what UAs it donates and accepts from, then
    find what shell labels these correspond to. Update the shells class
    instance with this information

    :param atom_idx: atom index to find HB labels for
    :param system: mdanalysis instance of all atoms in current frame
    :param HBs: HBCollection class instance
    :param shells: ShellCollection class instance
    """
    shell = shells.find_shell(atom_idx)
    accepts_from_labels = []
    donates_to_labels = []
    if shell:
        # 1.check if HB donating to and accepting from have previously
        # been found
        if shell.labels:
            # 2. Find what UAs the central atom donates to and accepts from
            # with the HBs class instance
            donates_to = HBs.find_acceptor(atom_idx)
            accepts_from = HBs.find_donators(atom_idx)
            if accepts_from:
                for d_idx in accepts_from:
                    # 3. check if UA being accepted from is in shell of
                    # central UA and add label to list
                    bonded_UA = Selections.find_bonded_heavy_atom(d_idx, system)
                    if bonded_UA.index in shell.UA_shell:
                        shell_idx = shell.UA_shell.index(bonded_UA.index)
                        accepts_from_labels.append(shell.labels[shell_idx])
                    else:
                        # don't add donating neighbour if not in central UA
                        # shell
                        continue

            if donates_to:
                # 4. iterate through acceptors and add labels
                for d_idx, a_idx in donates_to.items():
                    acceptor = system.atoms[a_idx]
                    if acceptor.mass < 1.1:
                        acceptor = Selections.find_bonded_heavy_atom(a_idx, system)
                    shell_idx = shell.UA_shell.index(acceptor.index)
                    donates_to_labels.append(shell.labels[shell_idx])
    # 5. Add labelled neighbours to shells instance
    shell.donates_to_labels = donates_to_labels
    shell.accepts_from_labels = accepts_from_labels



"""
Functions for transforming atomic positions and forces
"""

import numpy as np
from numpy import linalg as LA

import waterEntropy.maths.trig as Trig
import waterEntropy.utils.selections as Selections


def get_torques(
    molecule, center_of_mass: np.ndarray, rotation_axes: np.ndarray, MI_axis: np.ndarray
):
    """
    For a selection of atoms, use their positions and forces to get the
    torque (3,) for that selection of atoms. The positions are first translated
    to the center of mass, then the translated positions and forces are
    rotated to align with the chosen rotation axes. Lastly the torque
    is calculated from the cross product of the transformed positions and
    forces, which is subsequently divided by the sqaure root of the moment
    of inertia axis.

    :param coords: mdanalysis instance of atoms selected for positions
    :param forces: mdanalysis instance of atoms selected for forces
    :param center_of_mass: a (3,) array of the chosen center of mass
    :param rotation_axes: a (3,3) array to rotate forces along
    :param MI_axis: a (3,) array for the moment of inertia axis center
    """
    MI_axis_sqrt = np.sqrt(MI_axis)  # sqrt moi to weight torques
    translated_coords = molecule.positions - center_of_mass
    rotated_coords = np.tensordot(translated_coords, rotation_axes.T, axes=1)
    rotated_forces = np.tensordot(molecule.forces, rotation_axes.T, axes=1)
    cross_prod = np.cross(rotated_coords, rotated_forces)
    torque = np.sum(np.divide(cross_prod, MI_axis_sqrt), axis=0)

    return torque


def get_rotated_sum_forces(molecule, rotation_axes: np.ndarray):
    """
    Rotated the forces for a given seletion of atoms along a particular rotation
    axes (3,3)

    :param molecule: mdanalysis instance of molecule
    :param rotation_axes: a (3,3) array to rotate forces along
    """
    forces_summed = np.sum(molecule.forces, axis=0)
    rotated_sum_forces = np.tensordot(forces_summed, rotation_axes.T, axes=1)
    return rotated_sum_forces


def get_mass_weighted_forces(molecule, rotation_axes: np.ndarray):
    """
    For a given set of atoms, sum their forces and rotate these summed forces
    using the rotation axes (3,3)

    :param molecule: mdanalysis instance of molecule
    :param rotation_axes: a (3,3) array to rotate forces along
    """
    rotated_sum_forces = get_rotated_sum_forces(molecule, rotation_axes)
    mass_sqrt = np.sum(molecule.masses) ** 0.5
    mass_weighted_force = rotated_sum_forces / mass_sqrt
    return mass_weighted_force  # (3,)


def get_covariance_matrix(ft: np.ndarray, halve=0.5):
    """
    Get the outer product of the mass weighted forces or torques (ft) and
    half values if halve=True

    :param ft: (3,) array of either mass weighted forces or torques
    :param halve: Boolean to set weather covariance matrix should be halved
        (i.e. divide by :math:`2^2`)
    """
    cov_matrix = np.outer(ft, ft)
    if halve:
        cov_matrix *= halve**2
    return cov_matrix


def get_UA_masses(molecule):
    """
    For a given molecule, return a list of masses of UAs
    (combination of the heavy atoms + bonded hydrogen atoms. This list is used to
    get the moment of inertia tensor for molecules larger than one UA

    :param molecule: mdanalysis instance of molecule
    """
    UA_masses = []
    for atom in molecule:
        if atom.mass > 1.1:
            UA_mass = atom.mass
            bonded_atoms = molecule.select_atoms(f"bonded index {atom.index}")
            bonded_H_atoms = bonded_atoms.select_atoms("mass 1 to 1.1")
            for H in bonded_H_atoms:
                UA_mass += H.mass
            UA_masses.append(UA_mass)
        else:
            continue
    return UA_masses


def get_axes(molecule, molecule_scale: str):
    """
    From a selection of atoms, get the ordered principal axes (3,3) and
    the ordered moment of inertia axes (3,) for that selection of atoms

    :param molecule: mdanalysis instance of molecule
    :param molecule_scale: the length scale of molecule
    """
    # default moment of inertia
    moment_of_inertia = molecule.moment_of_inertia()
    if molecule_scale == "single_UA":
        pass  # moment_of_inertia = molecule.moment_of_inertia()
    if molecule_scale == "multiple_UAs":
        UAs = Selections.find_molecule_UAs(molecule)
        center_of_mass = molecule.center_of_mass()
        masses = get_UA_masses(molecule)
        moment_of_inertia = MOI(center_of_mass, UAs.positions, masses)
    principal_axes = molecule.principal_axes()
    # diagonalise moment of inertia tensor here
    # pylint: disable=unused-variable
    eigenvalues, _eigenvectors = LA.eig(moment_of_inertia)
    # principal axes = eigenvectors.T[order])
    # comment: could get principal axes from transformed eigenvectors
    #           but would need to sort out directions, so use MDAnalysis
    #           function instead

    # sort eigenvalues of moi tensor by largest to smallest magnitude
    order = abs(eigenvalues).argsort()[::-1]  # decending order
    # principal_axes = principal_axes[order] # PI already ordered correctly
    MOI_axis = eigenvalues[order]

    return principal_axes, MOI_axis


def MOI(CoM: np.ndarray, positions: np.ndarray, masses: list):
    """
    Use this function to calculate moment of inertia for cases where the
    mass list will contain masses of UAs rather than individual atoms and
    the postions will be those for the UAs only (excluding the H atoms
    coordinates).

    :param CoM: a (3,) array of the chosen center of mass
    :param positions: a (N,3) array of point positions
    :param masses: a (N,) list of point masses
    """
    I = np.zeros((3, 3))
    for coord, mass in zip(positions, masses):
        I[0][0] += (abs(coord[1] - CoM[1]) ** 2 + abs(coord[2] - CoM[2]) ** 2) * mass
        I[0][1] -= (coord[0] - CoM[0]) * (coord[1] - CoM[1]) * mass
        I[1][0] -= (coord[0] - CoM[0]) * (coord[1] - CoM[1]) * mass

        I[1][1] += (abs(coord[0] - CoM[0]) ** 2 + abs(coord[2] - CoM[2]) ** 2) * mass
        I[0][2] -= (coord[0] - CoM[0]) * (coord[2] - CoM[2]) * mass
        I[2][0] -= (coord[0] - CoM[0]) * (coord[2] - CoM[2]) * mass

        I[2][2] += (abs(coord[0] - CoM[0]) ** 2 + abs(coord[1] - CoM[1]) ** 2) * mass
        I[1][2] -= (coord[1] - CoM[1]) * (coord[2] - CoM[2]) * mass
        I[2][1] -= (coord[1] - CoM[1]) * (coord[2] - CoM[2]) * mass

    return I


def get_custom_axes(a: np.ndarray, b_list: list, c: np.ndarray, dimensions: np.ndarray):
    r"""
    For atoms a, b_list and c, calculate the axis to rotate forces around:

    - axis1: use the normalised vector ab as axis1. If there is more than one bonded
      heavy atom (HA), average over all the normalised vectors calculated from b_list
      and use this as axis1). b_list contains all the bonded heavy atom
      coordinates.

    - axis2: use cross product of normalised vector ac and axis1 as axis2.
      If there are more than two bonded heavy atoms, then use normalised vector
      b[0]c to cross product with axis1, this gives the axis perpendicular to
      axis1.

    - axis3: the cross product of axis1 and axis2, which is perpendicular to
      axis1 and axis2.

    :param a: central united-atom coordinates (3,)
    :param b_list: list of heavy bonded atom positions (3,N)
    :param c: atom coordinates of either a second heavy atom or a hydrogen atom
        if there are no other bonded heavy atoms in b_list (where N=1 in b_list)
        (3,)
    :param dimensions: dimensions of the simulation box (3,)

    ::

          a          1 = norm_ab
         / \         2 = |_ norm_ab and norm_ac (use bc if more than 2 HAs)
        /   \        3 = |_ 1 and 2
      b       c

    """
    axis1 = np.zeros(3)
    # average of all heavy atom covalent bond vectors for axis1
    for b in b_list:
        ab_vector = Trig.get_vector(a, b, dimensions)
        # scale vector with distance
        ab_dist = np.sqrt((ab_vector**2).sum(axis=-1))
        scaled_vector = np.divide(ab_vector, ab_dist)
        axis1 += scaled_vector  # ab_vector

    if len(b_list) > 2:
        ac_vector = Trig.get_vector(b_list[0], c, dimensions)
    else:
        ac_vector = Trig.get_vector(a, c, dimensions)
    ac_dist = np.sqrt((ac_vector**2).sum(axis=-1))
    ac_vector_norm = np.divide(ac_vector, ac_dist)

    if len(b_list) > 2:
        axis2 = np.cross(ac_vector_norm, axis1)
    else:
        axis2 = np.cross(axis1, ac_vector_norm)
    axis3 = np.cross(axis1, axis2)

    custom_axes = np.array((axis1, axis2, axis3))

    return custom_axes


def get_flipped_axes(positions, custom_axes, center_of_mass, dimensions):
    """
    For a given set of custom axes, ensure the axes are pointing in the
    correct direction wrt the heavy atom position and the chosen center
    of mass.
    """
    # sorting out PIaxes for MoI for UA fragment
    custom_axis = np.sum(custom_axes**2, axis=1)
    PIaxes = custom_axes / custom_axis**0.5

    # get dot product of Paxis1 and CoM->atom1 vect
    # will just be [0,0,0]
    RRaxis = Trig.get_vector(positions[0], center_of_mass, dimensions)
    # flip each Paxis if its pointing out of UA
    for i in range(3):
        dotProd1 = np.dot(PIaxes[i], RRaxis)
        PIaxes[i] = np.where(dotProd1 < 0, -PIaxes[i], PIaxes[i])

    return PIaxes


def get_custom_PI_MOI(molecule, custom_rotation_axes, center_of_mass, dimensions):
    """
    Get MOI tensor (PIaxes) and center point coordinates (custom_MI_axis)
    for UA level, where eigenvalues and vectors are not used.
    Note, positions and masses are provided separately as some cases
    require using the positions of heavy atoms only, but the masses of all
    atoms for a given selection of atoms.

    :param coords: MDAnalysis instance of molecule
    :param custom_rotation_axes: (3,3) arrray of rotation axes
    :param center_of_mass: (3,) center of mass for collection of atoms N
    :param masses: (N,) list of masses for collection of atoms, note this
        should be the same length as coords. If there are no hydrogens in
        the coords array, then the masses of these should be added to the
        heavy atom
    :param dimensions: (3,) array of system box dimensions.
    """
    # sorting out PIaxes for MoI for UA fragment
    custom_rotation_axes = get_flipped_axes(
        molecule.positions, custom_rotation_axes, center_of_mass, dimensions
    )
    translated_coords = molecule.positions - center_of_mass
    custom_MI_axis = np.zeros(3)
    for coord, mass in zip(translated_coords, molecule.masses):
        axis_component = np.sum(
            np.cross(custom_rotation_axes, coord) ** 2 * mass, axis=1
        )
        custom_MI_axis += axis_component

    return custom_rotation_axes, custom_MI_axis


def get_bonded_axes(system, atom, dimensions):
    """
    For a given united atom, find how to select bonded atoms to get the axes
    for rotating forces around. Few cases for choosing united atom axes:

    ::

        X -- H = bonded to one or more light atom

        X -- R = bonded to one heavy atom

        R -- X -- H = bonded to one heavy and one light atom

        R1 -- X -- R2 = bonded to two heavy atoms

        R1 -- X -- R2 = bonded to more than two heavy atoms
              |
              R3

    Note that axis2 is calculated by taking the cross product between axis1 and
    the vector chosen for each case, dependent on bonding:

    - case1: if all the bonded atoms are hydrogens, then just use the moment of
      inertia as all the axes.

    - case2: no axes required to rotate forces.

    - case3: use XR vector as axis1, vector XH to calculate axis2

    - case4: use vector XR1 as axis1, and XR2 to calculate axis2

    - case5: get the sum of all XR normalised vectors as axis1, then use vector
      R1R2 to calculate axis2

    """
    # check atom is a heavy atom
    if not atom.mass > 1.1:
        return None
    position_vector = atom.position
    custom_axes = None
    # find the heavy bonded atoms and light bonded atoms
    heavy_bonded, light_bonded = Selections.find_bonded_atoms(atom.index, system)
    UA_all = atom + heavy_bonded + light_bonded
    # now find which atoms to select to find the axes for rotating forces
    if len(heavy_bonded) == 2:
        custom_axes = get_custom_axes(
            atom.position,
            [heavy_bonded[0].position],
            heavy_bonded[1].position,
            dimensions,
        )
    if len(heavy_bonded) == 1 and len(light_bonded) >= 1:
        custom_axes = get_custom_axes(
            atom.position,
            [heavy_bonded[0].position],
            light_bonded[0].position,
            dimensions,
        )
    if len(heavy_bonded) > 2:
        custom_axes = get_custom_axes(
            atom.position, heavy_bonded.positions, heavy_bonded[1].position, dimensions
        )
    if len(heavy_bonded) == 1 and len(light_bonded) == 1:
        custom_axes = get_custom_axes(
            atom.position, [heavy_bonded[0].position], np.zeros(3), dimensions
        )
    if len(heavy_bonded) == 0:
        # !! Check if this scale is correct
        custom_axes, position_vector = get_axes(UA_all, molecule_scale="single_UA")

    if custom_axes is not None:
        custom_axes, position_vector = get_custom_PI_MOI(
            UA_all, custom_axes, atom.position, dimensions
        )

    return custom_axes, position_vector
